/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.util.Collection;

import org.junit.Assert;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;

/**
 * Simulation eines System Axl Datenendgeraets.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class AxlDeKanal255Sim extends AxlDeSim {

	/**
	 * Fehlercode fuer Negative Quittung.
	 */
	private static final int PUFFERINHALT_NICHTVERFUEGBAR = 12;
	/**
	 * Datenbeschreibung für Abrufe von AXL-DATEN.
	 */
	private static DataDescription ddAxlAbruf;
	/**
	 * Datenbeschreibung für Negative Quittungen.
	 */
	private static DataDescription ddAxlNegQuit;
	/**
	 * Ob man angemeldet ist fuer Abruf Telegramme.
	 */
	private boolean anmeldetAbrufPuffer;
	/**
	 * Ob fuer LveNegativeQuittung anmeldet ist.
	 */
	private boolean anmeldetAxlNegQuittung;
	/**
	 * Liste aller zuhoerigen DEs "broadcast Domene".
	 */
	private final Collection<AxlDeSim> listDes;
	/**
	 * Ob man negative Quittung senden soll.
	 */
	private boolean anfrageNichtVerstanden;

	/**
	 * Standardkonstruktor.
	 *
	 * @param dePid
	 *            Pid des DEs
	 * @param msPid
	 *            Pid der Axl MessStelle
	 * @param listDes
	 *            Liste der DEs die in der "broadcast Domene" liegen
	 */
	public AxlDeKanal255Sim(final String dePid, final String msPid, final Collection<AxlDeSim> listDes) {
		super(dePid, msPid, 1);
		this.listDes = listDes;
	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {
		super.initialize(connection);
		AxlDeKanal255Sim.ddAxlAbruf = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_AXLABRUF,
				AbfrPufferKonstanten.ASP_TLS_ABRUF, (short) 0);
		DeSim.getDav().subscribeReceiver(this, DeSim.getDav().getDataModel().getObject(getDePid()),
				AxlDeKanal255Sim.ddAxlAbruf, ReceiveOptions.normal(), ReceiverRole.receiver());
		anmeldetAbrufPuffer = true;

		// NegativeQuittung Senden
		AxlDeKanal255Sim.ddAxlNegQuit = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_AXLNEGATIVE_QUITTUNG,
				AbfrPufferKonstanten.ASP_TLS_ANTWORT, (short) 0);
		DeSim.getDav().subscribeSender(this, DeSim.getDav().getDataModel().getObject(getDePid()),
				AxlDeKanal255Sim.ddAxlNegQuit, SenderRole.source());
		anmeldetAxlNegQuittung = true;
	}

	@Override
	public void update(final ResultData[] results) {
		super.update(results);
		for (final ResultData resDatei : results) {
			final Data data = resDatei.getData();

			if ((data != null) && resDatei.getDataDescription().getAttributeGroup().getPid()
					.equals(AbfrPufferKonstanten.ATG_TLS_AXLABRUF)) {
				final long zeit = data.getTimeValue(AbfrPufferKonstanten.ZEITPUNKT).getMillis();
				// System.out.println("ZEIT: " + new java.util.Date(zeit));
				final int anzahl = data.getUnscaledValue(AbfrPufferKonstanten.ATR_ANZAHL).intValue();

				long aeltesteAbruf = AbfrPufferKonstanten.ZEIT_BEGINN;
				long kleinsteIntervall = AbfrPufferKonstanten.INTERVALL_MAX;

				for (final AxlDeSim de : listDes) {
					if (de.isOnline() && (de.getLetzterZeitStempellGeschickt() < aeltesteAbruf)) {
						aeltesteAbruf = de.getLetzterZeitStempellGeschickt();
					} else if (de.isOnline() && (de.getErzeugungIntervall() < kleinsteIntervall)) {
						kleinsteIntervall = de.getErzeugungIntervall();
					}
				}

				Assert.assertTrue("SysDe: Abruf fur Zeitpunkt das schon abgerufen ist", aeltesteAbruf < zeit);
				Assert.assertTrue("SysDe: Anzahl der DatenSaetze zu hoch",
						((ZeitSimulator.getTimeMillis() - zeit) / kleinsteIntervall) >= anzahl);

				if (anfrageNichtVerstanden) {
					sendeNegativeQuittung();
				} else {
					for (final AxlDeSim de : listDes) {
						de.sendeAchslastdaten(zeit, anzahl);
					}
				}
			}
		}
	}

	/**
	 * Sendet eine Negative Quittung.
	 */
	public void sendeNegativeQuittung() {
		final Data data = DeSim.getDav().createData(
				DeSim.getDav().getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_TLS_LVENEGATIVE_QUITTUNG));

		data.getUnscaledValue(AbfrPufferKonstanten.ATR_CODE).set(AxlDeKanal255Sim.PUFFERINHALT_NICHTVERFUEGBAR);
		data.getUnscaledValue(AbfrPufferKonstanten.ATR_HERSTELLER).set(0);
		data.getTextValue(AbfrPufferKonstanten.ATR_EINZELLTELEGRAMM).setText("Nicht verfuegbar");

		try {
			final ResultData resDatei = new ResultData(DeSim.getDav().getDataModel().getObject(getDePid()),
					AxlDeKanal255Sim.ddAxlNegQuit, ZeitSimulator.getTimeMillis(), data);
			DeSim.getDav().sendData(resDatei);
		} catch (final Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	@Override
	public void unsubscribe() {
		if (anmeldetAbrufPuffer) {
			DeSim.getDav().unsubscribeReceiver(this, DeSim.getDav().getDataModel().getObject(getDePid()),
					AxlDeKanal255Sim.ddAxlAbruf);
			anmeldetAbrufPuffer = false;
		}
		if (anmeldetAxlNegQuittung) {
			DeSim.getDav().unsubscribeReceiver(this, DeSim.getDav().getDataModel().getObject(getDePid()),
					AxlDeKanal255Sim.ddAxlNegQuit);

		}
		super.unsubscribe();
	}

	/**
	 * Setzt dem Flag, wenn der SystemDe immer mit negativer Quittung antwortet.
	 *
	 * @param anfrageNichtVerstanden
	 *            true, wenn negative quitungen geschickt werden sollen
	 */
	public void setAnfrageNichtVerstanden(final boolean anfrageNichtVerstanden) {
		this.anfrageNichtVerstanden = anfrageNichtVerstanden;
	}

}
