/*
 * Segment 4 Datenübernahme und Aufbereitung, SWE 4.5 Abfrage Puffer
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.abfrpuffer;

import java.util.Calendar;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAUtensilien;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;

/**
 * Simulation eines Axl Datenendgeraets.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 *
 */
public class AxlDeSim extends DeSim {

	/**
	 * Stringkonstanten.
	 */
	static final String VER_ERGEBNIS_MELDUNG = "DatenversionErgebnisMeldung";
	static final String ERFASSUNGS_INTERVALLDAUER = "Erfassungsintervalldauer";
	static final String VER_EINZELERGEBNIS_MELDUNG = "DatenversionEinzelErgebnisMeldung";
	static final String SPEICHERUNG_GERAET = "SpeicherungImGerät";

	/**
	 * Ob fuer AxlBetriebsParameter Daten anmeldet ist.
	 */
	private boolean anmeldetBetriebsParam;
	/**
	 * Ob fuer LangZeit VerkehrsDaten anmeldet ist.
	 */
	private boolean anmeldetAxlDaten;
	/**
	 * Datenbeschreibung für AXL-LZD.
	 */
	private static DataDescription ddAxlDatenSend;
	/**
	 * Datenbeschreibung von AXL-Betriesparameter.
	 */
	private static DataDescription ddAxlParam;
	/**
	 * Datenbeschreibung von AXL-NegativeQuittung.
	 */
	private static DataDescription ddAxlNegQuit;
	/**
	 * Der Pid der assoziierten Messtelle.
	 */
	private final String msPid;
	/**
	 * Erzeugungsintervall.
	 */
	private final long erzeugungIntervall;
	/**
	 * Der Betriebsparameter mit dem Erzeugungsintervall.
	 */
	private final long erzeugungIntBetriebParam;

	/**
	 * Standardkonstruktor.
	 *
	 * @param dePid
	 *            Pid des DEs
	 * @param msPid
	 *            Pid der Axl Messstelle
	 * @param erzeugungsIntervall
	 *            Erzeugungsintervall der LZD
	 */
	public AxlDeSim(final String dePid, final String msPid, final long erzeugungsIntervall) {
		super(dePid);
		this.msPid = msPid;
		this.erzeugungIntervall = erzeugungsIntervall * AbfrPufferKonstanten.STUNDE_IN_MS;
		this.erzeugungIntBetriebParam = AbfrPufferKonstanten.convertZeitZurAxlZeitCode(this.erzeugungIntervall);
		setLetzterZeitStempellGeschickt(AbfrPufferKonstanten.ZEIT_BEGINN);
	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {
		super.initialize(connection);
		// BetriebsParameter Senden
		AxlDeSim.ddAxlParam = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_AXLBETRIEBSPARAM,
				AbfrPufferKonstanten.ASP_PARAMETERSOLL, (short) 0);
		DeSim.getDav().subscribeSender(AxlDeSim.this, DeSim.getDav().getDataModel().getObject(getDePid()),
				AxlDeSim.ddAxlParam, SenderRole.source());
		anmeldetBetriebsParam = true;

		// Achslastdaten senden
		if (msPid != null) {
			AxlDeSim.ddAxlDatenSend = erzeugeDatenBeschreibung(AbfrPufferKonstanten.ATG_TLS_AXLERGEBNIS_MELDUNG,
					AbfrPufferKonstanten.ASP_TLS_ANTWORT, (short) 0);
		}
	}

	/**
	 * Sendet Achslastdaten, wenn keine Daten vorhanden sind, dann eine Negative
	 * Quittung.
	 *
	 * @param zeitStempel
	 *            ZeitStemepel des ersten Telegramms
	 * @param anzahl
	 *            Anzahl der Telegramms
	 */
	public void sendeAchslastdaten(final long zeitStempel, final int anzahl) {
		// if(msPid == null) return;

		final Data datensatz = DeSim.getDav().createData(
				DeSim.getDav().getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_TLS_AXLERGEBNIS_MELDUNG));
		setAxlDatenLeer(datensatz);
		datensatz.getTimeValue("T").setMillis(erzeugungIntervall);
		final ResultData resDatei = new ResultData(DeSim.getDav().getDataModel().getObject(msPid),
				AxlDeSim.ddAxlDatenSend, zeitStempel, datensatz, true);

		final long zeit = ZeitSimulator.getTimeMillis();

		// ////////////
		// JUnit Tests
		// ////////////
		// Assert.assertTrue("AxlDe: zeitStempel<zeit :" +
		// AbfrPufferKonstanten.datumFormat.format(new Date(zeitStempel))
		// + "<" + AbfrPufferKonstanten.datumFormat.format(new Date(zeit)) ,
		// zeitStempel<zeit);
		// Assert.assertTrue("AxlDe: Anzahl " + anzahl, anzahl < (zeit -
		// zeitStempel)/erzeugungIntervall);
		// Assert.assertTrue("AxlDe: Anfrage bekommen, obwohl nicht online bin",
		// online && !ohneAntwort);
		// Assert.assertTrue("AxlDe: DatenLuecke"+ new
		// Date(letzterZeitStempellGeschickt) + ", " +
		// (Math.abs(letzterZeitStempellGeschickt + erzeugungIntervall -
		// zeitStempel)), letzterZeitStempellGeschickt + erzeugungIntervall >
		// zeitStempel);

		int i = 0;
		for (i = 1; i < (anzahl + 1); i++) {
			try {
				resDatei.setDataTime((zeitStempel - (zeitStempel % erzeugungIntervall)) + (i * erzeugungIntervall));
				if (anmeldetAxlDaten) {
					DeSim.getDav().sendData(resDatei);
				} else if (msPid != null) {
					DeSim.getDav().subscribeSource(this, resDatei);
					anmeldetAxlDaten = true;
				}

				setLetzterZeitStempellGeschickt(
						(zeitStempel - (zeitStempel % erzeugungIntervall)) + (i * erzeugungIntervall));
			} catch (final Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}

		try {
			final ResultData resDatei2 = new ResultData(DeSim.getDav().getDataModel().getObject(msPid),
					AxlDeSim.ddAxlDatenSend, zeitStempel, null, false, DataState.NO_DATA);
			resDatei2.setDataTime((zeitStempel - (zeitStempel % erzeugungIntervall)) + (i * erzeugungIntervall) + 150);
			DeSim.getDav().sendData(resDatei2);
		} catch (final Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Initialisiert das Datensatz, damit es keine undefinierte Werte enthaltet.
	 *
	 * @param datensatz
	 */
	private void setAxlDatenLeer(final Data datensatz) {

		final String[] ueberladungen = new String[] { "AnzÜberladungLkw", "AnzÜberladungLkwA", "AnzÜberladungBusse" };
		final String[] anzGewichtsKlassen = new String[] { "AnzGewichtsKlasseLkw", "AnzGewichtsKlasseLkwA",
		"AnzGewichtsKlasseBusse" };

		DUAUtensilien.getAttributDatum("AnzEinzelachslastKlasse", datensatz).asArray().setLength(13);
		for (int j = 0; j < 13; j++) {
			DUAUtensilien.getAttributDatum("AnzEinzelachslastKlasse." + j, datensatz).asUnscaledValue().set(0);
		}

		DUAUtensilien.getAttributDatum("AnzDoppelachslastKlasse", datensatz).asArray().setLength(10);
		for (int j = 0; j < 10; j++) {
			DUAUtensilien.getAttributDatum("AnzDoppelachslastKlasse." + j, datensatz).asUnscaledValue().set(0);
		}

		DUAUtensilien.getAttributDatum("AnzDreifachachslastKlasse", datensatz).asArray().setLength(11);
		for (int j = 0; j < 11; j++) {
			DUAUtensilien.getAttributDatum("AnzDreifachachslastKlasse." + j, datensatz).asUnscaledValue().set(0);
		}

		for (final String s : ueberladungen) {
			DUAUtensilien.getAttributDatum(s, datensatz).asUnscaledValue().set(0);
		}

		for (final String element : anzGewichtsKlassen) {
			DUAUtensilien.getAttributDatum(element, datensatz).asArray().setLength(10);
			for (int j = 0; j < 10; j++) {
				DUAUtensilien.getAttributDatum(element + "." + j, datensatz).asUnscaledValue().set(0);
			}
		}
	}

	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		super.parseArguments(argumentList);
	}

	@Override
	public void update(final ResultData[] results) {
		super.update(results);
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {

		if (anmeldetBetriebsParam
				&& dataDescription.getAttributeGroup().getPid().equals(AbfrPufferKonstanten.ATG_TLS_AXLBETRIEBSPARAM)
				&& dataDescription.getAspect().getPid().equals(AbfrPufferKonstanten.ASP_PARAMETERSOLL)
				&& (state == ClientSenderInterface.START_SENDING)) {

			final Data data = DeSim.getDav().createData(
					DeSim.getDav().getDataModel().getAttributeGroup(AbfrPufferKonstanten.ATG_TLS_AXLBETRIEBSPARAM));

			data.getItem(AxlDeSim.VER_ERGEBNIS_MELDUNG).asUnscaledValue().set(10);
			data.getItem(AxlDeSim.ERFASSUNGS_INTERVALLDAUER).asUnscaledValue()
			.set((erzeugungIntervall / AbfrPufferKonstanten.STUNDE_IN_MS) * 4);
			data.getItem(AxlDeSim.VER_EINZELERGEBNIS_MELDUNG).asUnscaledValue().set(2);
			data.getItem(AxlDeSim.SPEICHERUNG_GERAET).asUnscaledValue().set(1);

			final ResultData resDatei = new ResultData(object, dataDescription,
					Calendar.getInstance().getTimeInMillis(), data);
			try {
				DeSim.getDav().sendData(resDatei);
			} catch (final Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		} else
			if (anmeldetBetriebsParam
					&& dataDescription.getAttributeGroup().getPid()
							.equals(AbfrPufferKonstanten.ATG_TLS_AXLBETRIEBSPARAM)
					&& dataDescription.getAspect().getPid().equals(AbfrPufferKonstanten.ASP_PARAMETERSOLL)
					&& (state == ClientSenderInterface.STOP_SENDING)) {
			final ResultData resDatei = new ResultData(object, dataDescription,
					Calendar.getInstance().getTimeInMillis(), null);
			try {
				DeSim.getDav().sendData(resDatei);
			} catch (final Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		} else {
			super.dataRequest(object, dataDescription, state);
		}
	}

	@Override
	public void unsubscribe() {
		if (anmeldetBetriebsParam) {
			DeSim.getDav().unsubscribeSender(AxlDeSim.this, DeSim.getDav().getDataModel().getObject(getDePid()),
					AxlDeSim.ddAxlParam);
			anmeldetBetriebsParam = false;
		}
		if (anmeldetAxlDaten) {
			DeSim.getDav().unsubscribeSender(AxlDeSim.this, DeSim.getDav().getDataModel().getObject(msPid),
					AxlDeSim.ddAxlDatenSend);
			anmeldetAxlDaten = false;
		}
		super.unsubscribe();
	}

	/**
	 * Liefert den Erzeugungsintervall.
	 *
	 * @return Erzeugungsintervall
	 */
	public long getErzeugungIntervall() {
		return erzeugungIntervall;
	}

	/**
	 * Liefert das Pid der Axl Messstelle.
	 *
	 * @return Pid
	 */
	public String getMsPid() {
		return msPid;
	}
}
